from collections import Counter
from typing import Any, ClassVar, Literal

import numpy as np
from bratly import EntityAnnotation, Fragment
from pydantic import BaseModel


class FragmentAgreement(BaseModel):
    """
    Object describing the similarity between two fragments (gold_frag, parallel_frag).
    Flag is correct if ends and starts are the same,
    partial if there is a non-empty intersection between the fragments, but either start or end or both differ.
    For partial 8 cases exist (see readme)
    Flag is spurious if there is no corresponding gold for parallel,
    missing if there is no corresponding parallel for gold.
    """

    gold_frag: Fragment
    parallel_frag: Fragment
    flag: int  # 0: correct, #1-8: partial 1-8, #9: spurious, #10: missing

    def __init__(
        self,
        gold_frag: Fragment | None = None,
        parallel_frag: Fragment | None = None,
    ) -> None:
        """
        Init: Either gold or parallel fragment have to be specified"""
        object.__setattr__(self, "parallel_frag", parallel_frag)
        object.__setattr__(self, "gold_frag", gold_frag)
        if not parallel_frag and not gold_frag:
            raise NotImplementedError
        if not gold_frag:
            object.__setattr__(self, "flag", 9)
        elif not parallel_frag:
            object.__setattr__(self, "flag", 10)
        else:
            if self.parallel_frag.start < self.gold_frag.start:
                flag = 6
            elif self.parallel_frag.start == self.gold_frag.start:
                flag = 0
            else:
                flag = 3
            if self.parallel_frag.end < self.gold_frag.end:
                flag += 1
            elif self.parallel_frag.end > self.gold_frag.end:
                flag += 2
            object.__setattr__(self, "flag", flag)

    def tag_from_flag(self) -> str:
        """
        Returns the str name of the flag stored as an integer"""
        if self.flag == 0:
            return "correct"
        if self.flag < 9:
            return f"partial{self.flag}"
        if self.flag == 9:
            return "spurious"
        return "missing"

    def contains(self, other: "FragmentAgreement") -> bool:
        """Returns whether the the scope of current agreement is larger then that of the other."""
        return (max(self.gold_frag.start, self.parallel_frag.start) <= max(other.gold_frag.start, other.parallel_frag.start)) and (
            min(self.gold_frag.end, self.parallel_frag.end) >= min(other.gold_frag.end, other.parallel_frag.end)
        )

    def is_distinct(self, others: list["FragmentAgreement"]) -> bool:
        """Returns whether it is distinct from any other agreements in the list."""
        return all(not o.contains(self) for o in others)

    def __str__(self) -> str:
        """For printing: returns the flag of the agreement and the two fragmentes compared"""
        return f"{self.tag_from_flag()}: {self.parallel_frag},{self.gold_frag}"

    def __repr__(self) -> str:
        return str(self)

    def __lt__(self, other: Any) -> bool:
        """
        Returns whether the scope of the current FragmentAgreement is occuring earlier than that of the others.
        The scope occurs earlier if it starts earlier, or - in case of same start - it ends earlier.
        For missing and spurious agreements, the scope is defined here as the scope of the missing or spurious fragment.
        """
        if isinstance(other, "FragmentAgreement"):
            comparison_vector = [-1, -1, -1, -1]
            if self.parallel_frag and self.gold_frag:
                comparison_vector[0:2] = [
                    max(self.parallel_frag.start, self.gold_frag.start),
                    min(self.parallel_frag.end, self.gold_frag.end),
                ]
            elif self.gold_frag:
                comparison_vector[0:2] = [self.gold_frag.start, self.gold_frag.end]
            elif self.parallel_frag:
                comparison_vector[0:2] = [
                    self.parallel_frag.start,
                    self.parallel_frag.end,
                ]
            else:
                raise NotImplementedError
            if other.parallel_frag and other.gold_frag:
                comparison_vector[2:4] = [
                    max(other.parallel_frag.start, other.gold_frag.start),
                    min(other.parallel_frag.end, other.gold_frag.end),
                ]
            elif other.gold_frag:
                comparison_vector[2:4] = [other.gold_frag.start, other.gold_frag.end]
            elif other.parallel_frag:
                comparison_vector[2:4] = [
                    other.parallel_frag.start,
                    other.parallel_frag.end,
                ]
            else:
                raise NotImplementedError
            return comparison_vector[0:2] < comparison_vector[2:4]
        return False


class Agreement(BaseModel):
    """
    Object describing the similarity between two entity annotations (gold, parallel).
    Flag is correct if ends and starts are the same and the two annotations have the same type;
    Incorrect if ends and starts are the same but the types are different;
    Partial if there is a non-empty intersection between the fragments, but either start or end or both differ and the type is the same,
    Related if the match is partial but the types are different.
    There are 2 cases for partial and for related flags, if there is partial/correct match between all fragments or when only between some of them.
    Flag is spurious if there is no corresponding gold for parallel,
    missing if there is no corresponding parallel for gold.
    """

    gold: EntityAnnotation
    parallel: EntityAnnotation
    fragment_agreements: list[FragmentAgreement]
    flag: int  # 0: correct, 1: incorrect, 2: missing, 3: spurious
    # 4: partial (all fragments), 5: partial (some fragments)
    # 6: related (all fragments), 7: related (some fragments)
    coverage: tuple[int, int]

    @staticmethod
    def merge_inclusive(fr1: Fragment, fr2: Fragment) -> Fragment:
        """Helper function returning the maximum scope (union) of two intersecting fragments"""
        if not fr1:
            return fr2
        if not fr2:
            return fr1
        return Fragment(min(fr1.start, fr2.start), max(fr1.end, fr2.end))

    @staticmethod
    def merge_exclusive(fr1: Fragment, fr2: Fragment) -> Fragment | None:
        """Helper function returning the minimum scope (intersection) of two intersecting fragments"""
        if not fr1 or not fr2 or fr1.end <= fr2.start or fr2.end <= fr1.start:
            return None
        return Fragment(max(fr1.start, fr2.start), min(fr1.end, fr2.end))

    @staticmethod
    def union(fragments: list[FragmentAgreement]) -> list[Fragment]:
        """Helper function returning the union of FragmentAgreements (i.e. eliminating duplicates)"""
        tmp = [
            Agreement.merge_inclusive(
                fragments[0].gold_frag,
                fragments[0].parallel_frag,
            ),
        ]
        j = 0
        for i, f in enumerate(sorted(fragments)[1:]):
            tmp2 = Agreement.merge_inclusive(
                fragments[i].gold_frag,
                fragments[i].parallel_frag,
            )
            if tmp[j] and tmp[j].end >= tmp2.start:
                tmp[j] = Fragment(tmp[j].start, tmp2.end)
            else:
                tmp.append([])
                j += 1
        return np.sum([f.end - f.start for f in tmp if f])

    @staticmethod
    def intersect(fragments: list[FragmentAgreement]) -> list[Fragment]:
        """Helper function returning the intersection of FragmentAgreements (contained by all of them)"""
        tmp = []
        for i, f in enumerate(sorted(fragments)[:-1]):
            tmp2 = Agreement.merge_exclusive(f, fragments[i + 1])
            if tmp2:
                tmp.append(tmp2)
        if tmp:
            return Agreement.union(tmp)
        return []

    def __init__(
        self,
        fragment_agreements: list[FragmentAgreement],
        parallel: EntityAnnotation | None,
        gold: EntityAnnotation | None,
    ) -> None:
        if len(fragment_agreements) < 1:
            raise NotImplementedError
        object.__setattr__(self, "gold", gold)
        object.__setattr__(self, "parallel", parallel)
        object.__setattr__(self, "fragment_agreements", fragment_agreements)
        object.__setattr__(
            self,
            "coverage",
            (
                Agreement.intersect(fragment_agreements),
                Agreement.union(fragment_agreements),
            ),
        )
        if all(fa.flag == 0 for fa in fragment_agreements):
            if len(fragment_agreements) == len(self.gold.fragments):
                # print(fragment_agreements,self.gold,self.parallel)
                if self.gold.label == self.parallel.label:
                    object.__setattr__(self, "flag", 0)
                else:
                    object.__setattr__(self, "flag", 1)
            elif self.gold.label == self.parallel.label:
                object.__setattr__(self, "flag", 5)
            else:
                object.__setattr__(self, "flag", 7)
        elif all(fa.flag == 9 for fa in fragment_agreements):
            object.__setattr__(self, "flag", 3)
        elif all(fa.flag == 10 for fa in fragment_agreements):
            object.__setattr__(self, "flag", 2)
        elif len(fragment_agreements) == len(self.gold.fragments):
            if not fragment_agreements[0].gold_frag or not fragment_agreements[0].parallel_frag:
                raise NotImplementedError
            if all(fa.flag < 9 and fa.flag > 0 for fa in fragment_agreements):
                if self.gold.label == self.parallel.label:
                    object.__setattr__(self, "flag", 4)
                else:
                    object.__setattr__(self, "flag", 6)
            elif self.gold.label == self.parallel.label:
                object.__setattr__(self, "flag", 5)
            else:
                object.__setattr__(self, "flag", 7)
        elif any(fa.flag < 9 and fa.flag > 0 for fa in fragment_agreements):
            if not fragment_agreements[0].gold_frag or not fragment_agreements[0].parallel_frag:
                raise NotImplementedError
            if self.gold.label == self.parallel.label:
                object.__setattr__(self, "flag", 5)
            else:
                object.__setattr__(self, "flag", 7)
        else:
            print(fragment_agreements, gold, parallel)
            raise NotImplementedError

    def flag_in_str(self) -> str:
        """Helper function converting the flag to the corresponding string"""
        match self.flag:
            case 0:
                return "CORRECT"
            case 1:
                return "INCORRECT"
            case 2:
                return "MISSING"
            case 3:
                return "SPURIOUS"
            case 4:
                return "PARTIAL_A"
            case 5:
                return "PARTIAL_S"
            case 6:
                return "RELATED_A"
            case 7:
                return "RELATED_S"
            case _:
                return "UNKNOWN"

    @staticmethod
    def annotation_to_csv(ann: EntityAnnotation, filename=""):
        """Converts one line of annotation to CSV format (part of one line)"""
        if ann:
            fr_str = ";".join([str(a) for a in ann.fragments])
            return f'{ann.id},{ann.label},{ann.get_start()},{ann.get_end()},{fr_str},"{ann.content}"'
        return ",,,,,"

    def to_csv(self, filename: str | None = None) -> str:
        """Converts an agreement to one line of CSV file"""
        flag_names = ["c", "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8", "s", "m"]
        interval_match = ";".join(
            [flag_names[fa.flag] for fa in self.fragment_agreements],
        )
        if filename:
            return f"{filename},{self.flag_in_str()},{self.annotation_to_csv(self.parallel)},{self.annotation_to_csv(self.gold)},{interval_match}\n"
        return f"{self.flag_in_str()},{self.annotation_to_csv(self.parallel)},{self.annotation_to_csv(self.gold)},{interval_match}\n"

    def __str__(self) -> str:
        return f"{self.flag_in_str()}\t{self.parallel}\t{self.gold}"


class MucTable(BaseModel):
    """
    Data structure object reflecting the parameters of MUC-5 scoring scheme.
    Extended with some additional parameters
    """

    agreements: list[Agreement]
    correct: int
    incorrect: int
    partial_A: int
    partial_S: int
    related_A: int
    related_S: int
    missing: int
    spurious: int
    # token / letter coverage of the annotations
    partial_match_coverage: tuple[int, int]

    RELAXED_COMPARISON: ClassVar[int] = 1
    STRICT_COMPARISON: ClassVar[int] = 2
    BORDER_COMPARISON: ClassVar[int] = 3

    @staticmethod
    def safe_divide(a: int, b: int)-> float | Literal[0]:  # If divident 0, return 0
        return 0 if b == 0 else a / b

    def __init__(self, agreements: list[Agreement]) -> None:
        object.__setattr__(self, "agreements", agreements)
        c = Counter([a.flag for a in self.agreements])
        # print("c", c, c[0], c[6])
        object.__setattr__(self, "correct", c[0])
        object.__setattr__(self, "incorrect", c[1])
        object.__setattr__(self, "missing", c[2])
        object.__setattr__(self, "spurious", c[3])
        object.__setattr__(self, "partial_A", c[4])
        object.__setattr__(self, "partial_S", c[5])
        object.__setattr__(self, "related_A", c[6])
        object.__setattr__(self, "related_S", c[7])
        object.__setattr__(
            self,
            "partial_match_coverage",
            (
                np.sum([a.coverage[0] for a in agreements]),
                np.sum([a.coverage[1] for a in agreements]),
            ),
        )

    def get_statistics(
        self,
        comparison_type: int = RELAXED_COMPARISON,
    ) -> dict[str, float]:
        """
        Returns the most importantant statistics values for the comparison
        as a dictionary.
        """
        stats = {}
        stats["CORRECT"] = np.sum(self.correct)
        stats["INCORRECT"] = np.sum(self.incorrect)
        stats["PARTIAL_ALLFRAGMENTS"] = np.sum(self.partial_A)
        stats["PARTIAL_SOMEFRAGMENTS"] = np.sum(self.partial_S)
        stats["PARTIAL"] = stats["PARTIAL_ALLFRAGMENTS"] + stats["PARTIAL_SOMEFRAGMENTS"]
        stats["RELATED_ALLFRAGMENTS"] = np.sum(self.related_A)
        stats["RELATED_SOMEFRAGMENTS"] = np.sum(self.related_S)
        stats["RELATED"] = stats["RELATED_ALLFRAGMENTS"] + stats["RELATED_SOMEFRAGMENTS"]
        stats["MISSING"] = np.sum(self.missing)
        stats["SPURIOUS"] = np.sum(self.spurious)
        stats["POSSIBLE"] = stats["CORRECT"] + stats["INCORRECT"] + stats["MISSING"] + stats["PARTIAL"]
        stats["ACTUAL"] = stats["CORRECT"] + stats["INCORRECT"] + stats["SPURIOUS"] + stats["PARTIAL"]
        stats["BORDER CORRECT"] = stats["CORRECT"] + stats["INCORRECT"]
        stats["INCORRECT_BORDER"] = stats["PARTIAL"] + stats["SPURIOUS"]
        stats["UNDERGENERATION"] = MucTable.safe_divide(
            stats["MISSING"],
            stats["POSSIBLE"],
        )
        stats["OVERGENERATION"] = MucTable.safe_divide(
            stats["SPURIOUS"],
            stats["ACTUAL"],
        )
        match comparison_type:
            case MucTable.RELAXED_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["CORRECT"] + stats["PARTIAL"] + stats["INCORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["CORRECT"] + stats["PARTIAL"] + stats["INCORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT"],
                    stats["CORRECT"] + stats["PARTIAL"] + stats["INCORRECT"],
                )
            case MucTable.STRICT_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["CORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["CORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT"] + stats["PARTIAL"],
                    stats["CORRECT"] + stats["PARTIAL"] + stats["INCORRECT"],
                )
            case MucTable.BORDER_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["BORDER CORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["BORDER CORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT_BORDER"],
                    stats["BORDER CORRECT"] + stats["INCORRECT_BORDER"],
                )
        if stats["PRECISION"] + stats["RECALL"] != 0:
            stats["F1"] = (2 * stats["PRECISION"] * stats["RECALL"])/(stats["PRECISION"] + stats["RECALL"])
        else:
            stats["F1"] = 0.0
        return stats


class MucCollection(BaseModel):
    """
    Class describing the comparison between several pairs of files.
    A collection of MucTables
    """

    muc_tables: list[MucTable]
    source_ids: list[str]
    corrects: list[int]
    incorrects: list[int]
    partial_As: list[int]
    partial_Ss: list[int]
    related_As: list[int]
    related_Ss: list[int]
    missings: list[int]
    spuriouses: list[int]

    def __init__(
        self,
        muc_tables: list[MucTable],
        source_ids: list[str] | None = None,
    ) -> None:
        object.__setattr__(self, "muc_tables", muc_tables)
        if not source_ids:
            object.__setattr__(self, "source_ids", ["" for _ in range(len(muc_tables))])
        object.__setattr__(self, "corrects", [mt.correct for mt in self.muc_tables])
        object.__setattr__(self, "incorrects", [mt.incorrect for mt in self.muc_tables])
        object.__setattr__(self, "partial_As", [mt.partial_A for mt in self.muc_tables])
        object.__setattr__(self, "partial_Ss", [mt.partial_S for mt in self.muc_tables])
        object.__setattr__(self, "related_As", [mt.related_A for mt in self.muc_tables])
        object.__setattr__(self, "related_Ss", [mt.related_S for mt in self.muc_tables])
        object.__setattr__(self, "missings", [mt.missing for mt in self.muc_tables])
        object.__setattr__(self, "spuriouses", [mt.spurious for mt in self.muc_tables])

    def get_global_statistics(
        self,
        comparison_type: int = MucTable.RELAXED_COMPARISON,
        with_help=False,
    ) -> dict[str, float]:
        object.__setattr__(self, "corrects", [mt.correct for mt in self.muc_tables])
        object.__setattr__(self, "incorrects", [mt.incorrect for mt in self.muc_tables])
        object.__setattr__(self, "partial_As", [mt.partial_A for mt in self.muc_tables])
        object.__setattr__(self, "partial_Ss", [mt.partial_S for mt in self.muc_tables])
        object.__setattr__(self, "related_As", [mt.related_A for mt in self.muc_tables])
        object.__setattr__(self, "related_Ss", [mt.related_S for mt in self.muc_tables])
        object.__setattr__(self, "missings", [mt.missing for mt in self.muc_tables])
        object.__setattr__(self, "spuriouses", [mt.spurious for mt in self.muc_tables])

        stats = {}
        stats["CORRECT"] = np.sum(self.corrects)
        stats["INCORRECT"] = np.sum(self.incorrects)
        stats["PARTIAL_ALLFRAGMENTS"] = np.sum(self.partial_As)
        stats["PARTIAL_SOMEFRAGMENTS"] = np.sum(self.partial_Ss)
        stats["PARTIAL"] = stats["PARTIAL_ALLFRAGMENTS"] + stats["PARTIAL_SOMEFRAGMENTS"]
        stats["RELATED_ALLFRAGMENTS"] = np.sum(self.related_As)
        stats["RELATED_SOMEFRAGMENTS"] = np.sum(self.related_Ss)
        stats["RELATED"] = stats["RELATED_ALLFRAGMENTS"] + stats["RELATED_SOMEFRAGMENTS"]
        stats["MISSING"] = np.sum(self.missings)
        stats["SPURIOUS"] = np.sum(self.spuriouses)
        stats["POSSIBLE"] = stats["CORRECT"] + stats["INCORRECT"] + stats["MISSING"] + stats["PARTIAL"] + stats["RELATED"]
        stats["ACTUAL"] = stats["CORRECT"] + stats["INCORRECT"] + stats["SPURIOUS"] + stats["PARTIAL"] + stats["RELATED"]
        stats["BORDER CORRECT"] = stats["CORRECT"] + stats["INCORRECT"]
        stats["INCORRECT_BORDER"] = stats["PARTIAL"] + stats["SPURIOUS"] + stats["RELATED"]
        stats["UNDERGENERATION"] = MucTable.safe_divide(
            stats["MISSING"],
            stats["POSSIBLE"],
        )
        stats["OVERGENERATION"] = MucTable.safe_divide(
            stats["SPURIOUS"],
            stats["ACTUAL"],
        )
        match comparison_type:
            case MucTable.RELAXED_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["CORRECT"] + stats["PARTIAL"] + stats["RELATED"] + stats["INCORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["CORRECT"] + stats["PARTIAL"] + stats["RELATED"] + stats["INCORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT"] + stats["PARTIAL"] + stats["RELATED"],
                    stats["CORRECT"] + stats["PARTIAL"] + stats["RELATED"] + stats["INCORRECT"],
                )
            case MucTable.STRICT_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["CORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["CORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT"],
                    stats["CORRECT"] + stats["PARTIAL"] + stats["RELATED"] + stats["INCORRECT"],
                )
            case MucTable.BORDER_COMPARISON:
                stats["RECALL"] = MucTable.safe_divide(
                    stats["BORDER CORRECT"],
                    stats["POSSIBLE"],
                )
                stats["PRECISION"] = MucTable.safe_divide(
                    stats["BORDER CORRECT"],
                    stats["ACTUAL"],
                )
                stats["SUBSTITUTION"] = MucTable.safe_divide(
                    stats["INCORRECT_BORDER"],
                    stats["BORDER CORRECT"] + stats["INCORRECT_BORDER"],
                )
        if stats["PRECISION"] + stats["RECALL"] != 0:
            stats["F1"] = (2 * stats["PRECISION"] * stats["RECALL"])/(stats["PRECISION"] + stats["RECALL"])
        else:
            stats["F1"] = 0.0
        if with_help:
            helping = "\nCORRECT = when annotation tags and indices match completely\n"
            helping += "INCORRECT = when annotation tags and attributes do not match, but the indices coincide\n"
            helping += "PARTIAL_ALLFRAGMENTS = when the annotation tags are the same but one of the annotations has the same end index and a different start index, for ALL fragments\n"
            helping += "PARTIAL_SOMEFRAGMENTS = same as previous, but for multi-fragment annotations with at least one CORRECT Fragment match\n"
            helping += "PARTIAL = PARTIAL_ALLFRAGMENTS + PARTIAL_SOMEFRAGMENTS\n"
            helping += "RELATED_ALLFRAGMENTS = when the annotation tags are different but one of the annotations has the same end index and a different start index, for ALL fragments\n"
            helping += "RELATED_SOMEFRAGMENTS = same as previous, but for multi-fragment annotations with at least one CORRECT Fragment match\n"
            helping += "RELATED = RELATED_ALLFRAGMENTS + RELATED_SOMEFRAGMENTS\n"
            helping += "MISSING = annotations existing only in the gold standard annotation set (aka. false negative)\n"
            helping += "SPURIOUS = annotations existing only in the candidate annotation set (aka. false positive)\n"
            helping += "\nPOSSIBLE = CORRECT + INCORRECT + MISSING + PARTIAL + RELATED\n"
            helping += "ACTUAL = CORRECT + INCORRECT + SPURIOUS + PARTIAL + RELATED\n"
            helping += "RECALL = (CORRECT + ~PARTIAL + ~RELATED + ~INCORRECT) / POSSIBLE\n"
            helping += "PRECISION = (CORRECT + ~PARTIAL + ~RELATED + ~INCORRECT) / ACTUAL\n"
            helping += "UNDERGENERATION = MISSING / POSSIBLE\n"
            helping += "OVERGENERATION = SPURIOUS / ACTUAL\n"
            helping += "SUBSTITUTION = (INCORRECT + ~PARTIAL + ~RELATED) / (CORRECT + PARTIAL + RELATED + INCORRECT) : (CORRECT + INCORRECT) / (CORRECT + INCORRECT + PARTIAL + SPURIOUS + RELATED) (for BORDER_COMPARISON)\n"
            helping += "F1-SCORE = 2 * PRECISION * RECALL / (PRECISION + RECALL)\n"
            helping += "BORDER CORRECT = CORRECT + INCORRECT\n"
            helping += "INCORRECT BORDER = PARTIAL + SPURIOUS + RELATED\n"
            helping += "\nCOMPARISON CONFIGURATIONS:\n"
            helping += "- RELAXED_COMPARISON = 1 (DEFAULT)\n"
            helping += "- STRICT_COMPARISON = 2\n"
            helping += "- BORDER_COMPARISON = 3\n"
            helping += "\n~PARTIAL and ~RELATED means it is considered only for RELAXED_COMPARISON config\n"
            helping += "~INCORRECT means it is considered only for BORDER_COMPARISON config\n"
            stats["Metrics HELP:"] = helping
        return stats



from pathlib import Path
import pandas as pd
from .count_tokens import count_tokens_and_anns
from ..agreement_types import compare_folders


def contains_any(main_string, substrings):
    return any([substring in main_string for substring in substrings])


def extract_diffs(csv: str, labels: list[str], fields: tuple, batch: dict, write_files: bool=True):
    concepts = {}
    for line in [line for line in csv.split("\n") if contains_any(line, labels)]:
        split_line = line.split(",")
        concept = (split_line[fields[0]], split_line[fields[1]])
        if concept not in concepts:
            concepts[concept] = { 'n':0, 'files':[]}
        concepts[concept]['n'] += 1
        concepts[concept]['files'].append(split_line[0])

    # sort concept by frequency
    sorted_concepts = sorted(concepts.items(), key=lambda item: item[1]['n'], reverse=True)

    # rearrange items from ((label,semantic_tag),n) to (label,semantic_tag,n)
    if 'bratpath' in batch:
        bratpath = (Path(batch['bratpath']) / batch['corrected'] / concept[1]['files'][0]).with_suffix('') # http://vmld-01500.huge.ad.hcuge.ch:8001/index.xhtml#/benchmark
    else:
        bratpath = ""

    sorted_concepts = [
        [concept[0][0].replace('"', ""), concept[0][1], concept[1]['n'], "", str(bratpath)]
        for concept in sorted_concepts
    ]

    # fileout
    fileout = batch["datapathout"] / f"compare_{batch['name']}_{'_'.join(labels)}.xlsx"

    # create dataframe
    df = pd.DataFrame(sorted_concepts, columns=["label", "semantic_tag", "count", "SCT_multicode", "example"])

    # df.to_excel(fileout,index=False)

    writer = pd.ExcelWriter(fileout, engine='xlsxwriter')
    df.to_excel(writer, sheet_name='Sheet1', index=False)
    workbook  = writer.book
    worksheet = writer.sheets['Sheet1']

    # Iterate through the DataFrame and write URLs as hyperlinks
    for row_num, example in enumerate(df['example'], start=1):
        worksheet.write_url(row_num, 4, example, string="/".join(example.split("/")[-3:]))
    writer.close()


    return len(df)


def compare_batch(batch: dict, write_files: bool = True, max_n: int = -1):
    """ both folders must be flat"""
    if "datapath" not in batch:
        exit("no datapath in batch dict")
    if "datapathout" not in batch:
        batch["datapathout"] = batch["datapath"]
    batch["datapath"] = Path(batch["datapath"])
    batch["datapathout"] = Path(batch["datapathout"])

    print("===============\nCompare folders")
    mc, csv, df = compare_folders(
        str(batch["datapath"] / batch["auto"]),
        str(batch["datapath"] / batch["corrected"]),
        max_n=max_n,
    )

    stats = mc.get_global_statistics()
    # print(batch["corrected"], "vs", batch["auto"], f"{_stats['POSSIBLE']}:{_stats['F1']:.2f} ({_stats['PRECISION']:.2f})")
    # pprint(stats)

    print("=================================\nCount tokens and annonated tokens")
    n_doc, n_tokens, n_ann_tokens = count_tokens_and_anns(
        batch["datapath"] / batch["auto"],
        max_n=max_n,
    )

    CSVOUT = Path(batch["datapathout"]) / f"compare_{batch['name']}.csv"

    if write_files:
        # with CSVOUT.open("w", encoding="utf_8") as f:
        #     f.write(csv)
        # print(df)
        df.to_excel(CSVOUT.with_suffix(".xlsx"), index=False)

    # create files
    n_unique_missing = extract_diffs(csv, ["MISSING"], (13, 9), batch, write_files=write_files)
    n_unique_partial = extract_diffs(csv, ["PARTIAL"], (13, 9), batch, write_files=write_files)
    n_unique_spurious = extract_diffs(csv, ["SPURIOUS"], (7, 3), batch, write_files=write_files)

    print("docs:", n_doc, "\ntokens", n_tokens, "\nannotations", n_ann_tokens)
    stat_list = [n_doc, n_tokens, n_ann_tokens,
                100 * n_ann_tokens / n_tokens,
                -1,                          # placeholder
                100 * stats["OVERGENERATION"],
                stats["SPURIOUS"],
                100 * stats["UNDERGENERATION"],
                stats["MISSING"],
                100 * stats["RECALL"],
                100 * stats["PRECISION"],
                100 * stats["F1"],
                n_unique_spurious,
                -1,                          # placeholder
                n_unique_missing,
                n_unique_partial
            ]
    print(";".join(f"{num:.2f}" if isinstance(num,float) else str(num) for num in stat_list))
    print("ndocs,ntoks,nanns,nanns/ntoks,-1,overgen,spurious,undergen,missing,recall,precision,f1,unique_spurious,-1,unique_missing,unique_partial")



import os
from itertools import compress
from typing import List, Optional, cast

import numpy as np
import pandas as pd
from bratly import AnnotationCollection, Document, DocumentCollection, EntityAnnotation, Fragment
from bratly_io_fs import (
    copy_txt_from_collection_to_another_path,
    parse_ann_file,
    read_ann_files_from_folder,
    read_document_collection_from_folder,
    write_ann_files_in_folder,
)

from ..agreement_types import *
from .count_tokens import count_tokens_and_anns, stat_ann_categories


def compare_annotations(
    annotation_file: AnnotationCollection,
    gold_standard_file: AnnotationCollection,
    filename: str,
    comparison_style="Naive",
    originalFile=None,
) -> tuple[MucTable, str]:
    """Compares two annotation files and returns
    * the MucTable describing the similarity between the two files
    * and a string containing the CSV lines corresponding to the comparison of files.
    Several comparison styles are to be implemented. Currently only Naive style is supported.
    * Naive: Includes every match (when there is at least 1 charcter overlap between two annotations)
    and any spurious and missing annotation.
    * Clean: Keeps only the most likely match for every annotations in the gold standard.
    * Largest: Keeps the largest scope of agreement assuming that
        annotations of the same sepearated by only whitespace in the original file might be joined.
    * Brat: Uses the definitions of the Brat standard (no related class, partial only if end matches)
    """
    annotation_entities: List[EntityAnnotation] = cast(List[EntityAnnotation], annotation_file.get_annotations(descendant_type=EntityAnnotation))
    gold_standard_entities: List[EntityAnnotation] = cast(List[EntityAnnotation], gold_standard_file.get_annotations(descendant_type=EntityAnnotation))

    if not comparison_style in ["Naive", "Clean", "Largest", "Brat"]:
        raise NotImplementedError
    if comparison_style == "Largest" and not isinstance(originalFile, str):
        raise NotImplementedError

    # Getting all fragments with a tuple key (annotation number, fragment number)
    af1 = sorted(
        [
            (m, n, f)
            for m, a in enumerate(annotation_entities)
            for n, f in enumerate(a.fragments)
        ],
        key=lambda x: x[2],
    )
    af2 = sorted(
        [
            (m, n, f)
            for m, a in enumerate(gold_standard_entities)
            for n, f in enumerate(a.fragments)
        ],
        key=lambda x: x[2],
    )
    if not af1 and not af2:
        return MucTable([]), ""
    fragment_agreements1 = [[] for _ in range(len(af1))]
    fragment_agreements2 = [[] for _ in range(len(af2))]
    af2_starts = [a[2].start for a in af2]
    status_af2 = [True for _ in range(len(af2))]
    status_untouched = [True for _ in range(len(af2))]
    ns = list(range(len(af2)))
    ms = list(range(len(af1)))
    j = 0

    # For all fragments in the parallel file find any potentially related gold fragments
    for i, ann1 in enumerate(af1):
        a1 = ann1[2]
        # Restrain search for gold fragments which start before the end of the current fragm
        # but doesn't end before the end of previously processed.
        k = np.searchsorted(af2_starts, a1.end, side="left")
        potential_annotations = list(compress(af2[j:k], status_af2[j:k]))
        potential_indices = list(compress(ns[j:k], status_af2[j:k]))
        if len(potential_annotations) < 1:
            fragment_agreements1[i].append((ann1[0], -1, FragmentAgreement(None, a1)))
        for l, ann2 in enumerate(potential_annotations):
            a2 = ann2[2]
            # If the gold fragment ends before this one
            if a2.end < a1.start:
                if len(fragment_agreements2[potential_indices[l]]) < 1:
                    fragment_agreements2[potential_indices[l]].append(
                        (-1, ann2[0], FragmentAgreement(a2, None))
                    )
                status_af2[potential_indices[l]] = False
            # If they intersect
            else:
                agreement = (ann1[0], ann2[0], FragmentAgreement(a2, a1))
                fragment_agreements2[potential_indices[l]].append(agreement)
                fragment_agreements1[i].append(agreement)
                status_untouched[potential_indices[l]] = False
        # If all gold fragments ends before this one
        if not fragment_agreements1[i]:
            fragment_agreements1[i].append((ann1[0], -1, FragmentAgreement(None, a1)))

        # Constrain for the next iteration
        j = np.searchsorted(status_af2, True)

    # Add missing annotation at the end of the file
    for i, ann2 in enumerate(af2[j : len(status_af2)]):
        if len(fragment_agreements2[i + j]) < 1:
            fragment_agreements2[i + j].append(
                (-1, ann2[0], FragmentAgreement(ann2[2], None))
            )

    tmp = []
    # Sort FragmentAgreements according the start and end of the agreement
    tmp1 = sorted(
        [fa for sub in fragment_agreements1 for fa in sub], key=lambda x: (x[0], x[1])
    )

    fas = [fa[2] for fa in tmp1]
    last_different = 0
    gold_not_missing = set()
    for i, a in enumerate(tmp1[:-1]):
        # If the next fragment agreement doesn't concern the same gold and parallel fragments
        if tmp1[i][0] != tmp1[i + 1][0] or tmp1[i][1] != tmp1[i + 1][1]:
            # not spurious
            if tmp1[i][1] != -1:
                tmp.append(
                    Agreement(
                        fas[last_different : i + 1],
                        annotation_entities[tmp1[i][0]],
                        gold_standard_entities[tmp1[i][1]],
                    )
                )
            # spurious
            else:
                tmp.append(
                    Agreement(
                        fas[last_different : i + 1],
                        annotation_entities[tmp1[i][0]],
                        None,
                    )
                )
            last_different = i + 1
            gold_not_missing.add(tmp1[i][1])
    # print('LS', last_different, tmp1, tmp1[last_different][1])
    # Process the remaining fragment agrements
    if len(tmp1) > 0:
        if tmp1[last_different][1] != None and tmp1[last_different][1] > -1:
            tmp.append(
                Agreement(
                    fas[last_different : len(tmp1)],
                    annotation_entities[tmp1[last_different][0]],
                    gold_standard_entities[tmp1[last_different][1]],
                )
            )
        else:
            tmp.append(
                Agreement(
                    fas[last_different : len(tmp1)],
                    annotation_entities[tmp1[last_different][0]],
                    None,
                )
            )
        gold_not_missing.add(tmp1[last_different][1])
    tmp2 = sorted(
        [fa for sub in fragment_agreements2 for fa in sub], key=lambda x: (x[1], x[0])
    )

    fas = [fa[2] for fa in tmp2]
    last_different = 0
    for i, a in enumerate(tmp2[:-1]):
        if tmp2[i][1] != tmp2[i + 1][1]:
            if not tmp2[i][1] in gold_not_missing:
                tmp.append(
                    Agreement(
                        fas[last_different : i + 1],
                        None,
                        gold_standard_entities[tmp2[i][1]],
                    )
                )
            last_different = i + 1

    if len(tmp2) > 0:
        if not tmp2[-1][1] in gold_not_missing:
            tmp.append(
                Agreement(
                    fas[last_different:], None, gold_standard_entities[tmp2[-1][1]]
                )
            )
    csv = "".join([c.to_csv(filename=filename) for c in tmp])
    mt = MucTable(tmp)
    # print(mt.get_statistics())
    return (mt, csv)


def write_results(
    eval_folder: str = "",
    eval_file: str = "",
    df_agreement: Optional[pd.DataFrame] = None,
    csvc: str = "",
    muc_coll: Optional[MucCollection] = None,
    comparison_type=MucTable.RELAXED_COMPARISON,
    verbose: bool = False,
    muc_with_help: bool = False,
):
    if eval_folder != "":
        if not os.path.exists(eval_folder):
            os.makedirs(eval_folder)
        assert(type(df_agreement) == pd.DataFrame)
        df_agreement.to_excel(
            os.path.join(eval_folder, eval_file + "_agreement_report.xlsx")
        )

        with open(
            os.path.join(eval_folder, eval_file + "_agreement_report.csv"),
            "w",
            encoding="utf_8",
        ) as f:
            f.write(csvc)

        # Generate the MUC (Message Understanding Conferance) Table (ref: https://github.com/savkov/bratutils)
        print("writing muc table"+ os.path.join(eval_folder, eval_file + "_muc_table.txt"))
        with open(
            os.path.join(eval_folder, eval_file + "_muc_table.txt"),
            "w",
            encoding="utf_8",
        ) as f:
            assert(type(muc_coll) == MucCollection)
            dico_stats = muc_coll.get_global_statistics(
                comparison_type=comparison_type, with_help=muc_with_help
            )
            for k, v in dico_stats.items():
                result = str(k) + "-" + str(v) + "\n"
                f.write(result)


COLUMNS = "filename,eval_tag,parallel_id,parallel_label,parallel_start,parallel_end,parallel_fragments,parallel_content,gold_id,gold_label,gold_start,gold_end,gold_fragments,gold_content,interval_match"
N_COLUMNS = len(COLUMNS.split(","))


def compare_folders(
    parallel_folder: str,
    gold_folder: str,
    header: bool = True,
    eval_folder: str = "",
    eval_file: str = "",
    muc_with_help: bool = False,
    keep_specific_annotations: Optional[list[str]] = None,
    comparison_type=MucTable.RELAXED_COMPARISON,
    verbose: bool = False,
) -> tuple[MucCollection, str, pd.DataFrame]:
    """Compares annotation files placed in two parallel folder.
    The function compares files with the same name and returns a MucCollection
    and the content of all agreements in CSV format."""
    parallel_texts = read_ann_files_from_folder(parallel_folder)
    gold_texts = read_ann_files_from_folder(gold_folder)
    muc_coll = MucCollection([])
    csvc = COLUMNS + "\n" if header else ""
    df_agreement = pd.DataFrame(columns=COLUMNS.split(","))

    for fname, text in parallel_texts.items():
        # print(fname, parse_textfile(gold_texts[fname]))
        if verbose:
            print("fname", fname)
            print("text", text)
        if fname in gold_texts:
            # print(fname)
            anns_text = parse_ann_file(text, fname)
            anns_gold = parse_ann_file(gold_texts[fname], fname)

            if keep_specific_annotations is not None:
                anns_text.keep_specific_annotations(keep_specific_annotations)
                anns_gold.keep_specific_annotations(keep_specific_annotations)

            mt, csv = compare_annotations(anns_text, anns_gold, fname)
            muc_coll.muc_tables.append(mt)
            csvc += csv
            # print(csv)

            for line in csv.split("\n"):
                line_split = line.split(",")
                if len(line_split) == N_COLUMNS:
                    df_agreement.loc[len(df_agreement)] = line_split
                else:
                    pass
                #print("wrong number of columns" + str(len(line_split)) + line)

    write_results(
        eval_folder=eval_folder,
        eval_file=eval_file,
        df_agreement=df_agreement,
        csvc=csvc,
        muc_coll=muc_coll,
        comparison_type=comparison_type,
        verbose=verbose,
        muc_with_help=muc_with_help,
    )

    return (muc_coll, csvc, df_agreement)


def compare_pairs_of_folders(
    pairs: List[tuple[str, str]],  # parallel, gold
    header: bool = True,
    eval_folder: str = "",
    eval_file: str = "",
    muc_with_help: bool = False,
    keep_specific_annotations: Optional[list[str]] = None,
    comparison_type=MucTable.RELAXED_COMPARISON,
    verbose: bool = False,
) -> tuple[MucCollection, str, pd.DataFrame]|None:

    global_muc_coll = MucCollection([])
    global_csvc = COLUMNS + "\n" if header else ""
    global_df_agreement = pd.DataFrame(columns=COLUMNS.split(","))

    for pair in pairs:
        print("Comparing", str(pair))
        muc_coll, csvc, df_agreement = compare_folders(
            pair[0],
            pair[1],
            header=False,
            eval_folder="",
            eval_file=eval_file,
            muc_with_help=muc_with_help,
            keep_specific_annotations=keep_specific_annotations,
            comparison_type=comparison_type,
            verbose=verbose,
        )

        global_muc_coll.muc_tables += muc_coll.muc_tables
        global_csvc += csvc
        global_df_agreement = pd.concat([global_df_agreement, df_agreement])


    # print(global_muc_coll)
    # print(global_muc_coll.muc_tables)
    write_results(
        eval_folder=eval_folder,
        eval_file=eval_file,
        df_agreement=global_df_agreement,
        muc_coll=global_muc_coll,
        csvc=global_csvc,
        comparison_type=comparison_type,
        verbose=verbose,
        muc_with_help=muc_with_help,
    )


def compare_many_folders(
    folders: list[str],
    basefolder: str = "",
    header: bool = True,
    eval_folder: str = "",
    muc_with_help: bool = False,
):
    """
    do 3 comparisons
    == stats by entity types
    == F-scores
    == error types compared to gold
    """

    strout = basefolder
    strout += str(folders) + "\n"
    strout += "\n== stats by entity types\n"
    print("== stats by entity types")
    strout += count_entities_of_many_folders(folders, basefolder=basefolder)

    mc = {}
    csv = {}
    df = {}
    stats = {}

    for i in range(1, len(folders)):
        for j in range(i, len(folders)):
            f1 = os.path.join(basefolder, folders[i - 1])
            f2 = os.path.join(basefolder, folders[j])
            strout += f"comparing {f1} {f2}" + "\n"
            print(f"comparing {f1} {f2}")

            mc[f"{i-1}_{j}"], csv[f"{i-1}_{j}"], df[f"{i-1}_j"] = compare_folders(
                f1,
                f2,
                header=header,
                eval_folder=eval_folder,
                eval_file=f"{folders[i-1]}_{folders[j]}",
                muc_with_help=muc_with_help,
            )
            stats[f"{i-1}_{j}"] = mc[f"{i-1}_{j}"].get_global_statistics()

    strout += "\n== F-scores\n"
    print("== F-scores")
    for j in range(1, len(folders)):
        strout += "\t" + folders[j]
    strout += "\n"
    for i in range(1, len(folders)):
        strout += folders[i - 1]
        strout += "\t" * (i - 1)
        for j in range(i, len(folders)):
            if f"{i-1}_{j}" in stats:
                strout += f"\t{stats[f'{i-1}_{j}']['F1']:.2f}"
        strout += "\n"
    strout += "\n"

    print("== error types compared to gold: ", folders[0])
    strout += "\n== error types compared to gold: " + folders[0] + "\n"
    strout += "\t" + folders[0] + " vs.\t" + "\t".join(folders[1:]) + "\n"

    corrects = [str(sum(mc[f"{0}_{i}"].corrects)) for i in range(1, len(folders))]
    strout += "CORRECT\t\t" + "\t".join(corrects) + "\n"

    incorrects = [str(sum(mc[f"{0}_{i}"].incorrects)) for i in range(1, len(folders))]
    strout += "INCORRECT\t" + "\t".join(incorrects) + "\n"

    partials = [
        str(sum(mc[f"{0}_{i}"].partial_As + mc[f"{0}_{i}"].partial_Ss))
        for i in range(1, len(folders))
    ]
    strout += "PARTIAL\t\t" + "\t".join(partials) + "\n"

    missings = [str(sum(mc[f"{0}_{i}"].missings)) for i in range(1, len(folders))]
    strout += "MISSING\t\t" + "\t".join(missings) + "\n"

    spuriouses = [str(sum(mc[f"{0}_{i}"].spuriouses)) for i in range(1, len(folders))]
    strout += "SPURIOUS\t" + "\t".join(spuriouses) + "\n"

    return strout


def count_entities_of_folder(folder, basefolder: str = "") -> tuple[dict, int, int]:
    types = {}
    n_file = 0
    n_ann = 0
    folder = os.path.join(basefolder, folder)

    # print(folder)
    dc = read_document_collection_from_folder(folder)
    assert(type(dc) == DocumentCollection)
    for doc in dc.documents:
        n_file += 1
        # print(doc.filename_without_ext)
        for anncoll in doc.annotation_collections:
            for ann in anncoll.annotations:
                if isinstance(ann, EntityAnnotation):
                    n_ann += 1
                    if ann.label not in types:
                        types[ann.label] = 0
                    types[ann.label] += 1
    return types, n_file, n_ann


def count_entities_of_many_folders(folders: list[str], basefolder: str = "") -> str:
    types: dict = {}
    type_list: list = []
    n_file = {}
    n_ann = {}
    for folder in folders:
        # print(folder)
        types[folder], n_file[folder], n_ann[folder] = count_entities_of_folder(
            folder, basefolder=basefolder
        )
        # print(types[folder].keys())
        type_list.extend(list(types[folder].keys()))
        # print(type_list)

    type_list = list(set(type_list))
    strout = ""

    for folder in folders:
        strout += "\t" + folder
    strout += "\n"
    for folder in folders:
        strout += f"\t{n_file[folder]}"
    strout += "\tfiles" + "\n"
    for folder in folders:
        strout += f"\t{n_ann[folder]}"
    strout += "\tannotations\n"

    for tlist_element in type_list:
        for folder in folders:
            if tlist_element in types[folder]:
                strout += "\t" + str(types[folder][tlist_element])
            else:
                strout += "\t0"
        strout += "\t" + tlist_element
        strout += "\n"

    return strout


def create_union_gold_candidate_sets(
    path_doc_col_gs: str,
    path_doc_col_auto: str,
    path_output_newgold_folder: str = "./new_goldstandard_to_fix/",
    copy_txt_files: bool = True,
) -> None:
    """CAREFUL: the changes is inplace - the folder given in input will see its annotations altered"""
    try:
        os.makedirs(path_output_newgold_folder)
    except OSError:
        print(
            "Security: The output folder already exists! Please, delete the folder, or change the output path."
        )
        return None

    # read doc cols
    colgs: DocumentCollection = cast(DocumentCollection, read_document_collection_from_folder(path_doc_col_gs))
    colcs: DocumentCollection = cast(DocumentCollection, read_document_collection_from_folder(path_doc_col_auto))

    col_new: DocumentCollection = DocumentCollection(path_output_newgold_folder)

    for docgs in colgs.documents:
        # for each docgs, find the actual doccs
        for doccs in colcs.documents:
            if doccs.filename_without_ext != docgs.filename_without_ext:
                continue
            # now, we know the docs correspond to each other
            annotcol_gs, annotcol_cs = (
                docgs.annotation_collections[0],
                doccs.annotation_collections[0],
            )
            # get annotations
            entitesauto: list[EntityAnnotation] = cast(list[EntityAnnotation], annotcol_cs.get_annotations(descendant_type=EntityAnnotation))
            entitesgs: list[EntityAnnotation] = cast(list[EntityAnnotation], annotcol_gs.get_annotations(descendant_type=EntityAnnotation))

            # find the CORRECT annotations, add _both suffix
            for elt1 in entitesauto:
                for elt2 in entitesgs:
                    if (
                        elt1.fragments == elt2.fragments and elt1.label == elt2.label
                    ):  # same tag, and same fragments
                        elt1.label = elt1.label + "_both"
                        elt2.label = elt2.label + "_both"

            # get the labels
            labels = []
            for elt in entitesauto:
                if elt.label in labels or elt.label.endswith("_both"):
                    continue
                else:
                    labels.append(elt.label)
            for elt in entitesgs:
                if elt.label in labels or elt.label.endswith("_both"):
                    continue
                else:
                    labels.append(elt.label)

            # add the suffixes
            for elt in labels:
                annotcol_gs.replace_annotation_labels(
                    old_name=elt, new_name=elt + "_gold", specific_type=EntityAnnotation
                )
                annotcol_cs.replace_annotation_labels(
                    old_name=elt, new_name=elt + "_auto", specific_type=EntityAnnotation
                )

            # we create a new annotcol which is the union of both
            annotcol_gs.combine(annotcol_cs, with_renum=True)
            annotcol_gs.remove_duplicates()  # duplicates are the CORRECT (_both suffix)
            annotcol_gs.renum()  # renum

            # new doc to add
            doc_new = Document(
                os.path.join(
                    path_output_newgold_folder, docgs.filename_without_ext + ".txt"
                ),
                annotation_collections=[annotcol_gs],
            )
            print(doc_new)
            col_new.add_document(doc_new)

    # write new corpus
    print(col_new)
    print(f"Writing Union collection in path:{path_output_newgold_folder}")
    write_ann_files_in_folder(col_new, path_output_newgold_folder)

    # also writing the txt file
    if copy_txt_files is True:
        print(
            f"Also writing the txt files, taken from gold, to path:{path_output_newgold_folder}"
        )
        copy_txt_from_collection_to_another_path(colgs, path_output_newgold_folder)



"""
This module provides functions to count tokens and annotations in document collections and to collect statistics about annotated categories.

Functions:
    - count_tokens_and_anns(folder: Path) -> tuple:
        Counts the total number of tokens and annotated tokens in a document collection.

    - stat_ann_categories(paths: list[Path|str], stats: dict = None, depth: int = 0) -> dict:
        Collects and prints statistics about annotated categories in documents, including the number of tokens, annotated tokens, annotations, and documents, as well as the occurrences of each annotation type.
"""

from pathlib import Path

import pandas as pd
from bratly import DocumentCollection, EntityAnnotation
from bratly_io_fs import read_document_collection_from_folder
from tqdm import tqdm


def count_tokens_and_anns(folder: Path):
    """
    Count the number of tokens and annotated tokens in documents within a specified folder.

    Args:
        folder (Path): The path to the folder containing the documents.

    Returns:
        tuple: A tuple containing:
            - int: The number of documents.
            - int: The total number of tokens.
            - int: The total number of annotated tokens.

    """
    n_ann_tokens, n_tokens = 0, 0
    doc_coll: DocumentCollection | None = read_document_collection_from_folder(str(folder))
    if doc_coll is None:
        return 0, 0, 0
    print("counting tokens and annotated token")

    doc_coll.replace_annotation_labels("", "label", specific_type=None, all_labels=True)
    for i_doc, doc in enumerate(doc_coll.documents):
        print(i_doc + 1, "/", len(doc_coll.documents), end="\r")
        doc_coll.documents[i_doc].annotation_collections[0].remove_contained_annotations()
        doc_coll.documents[i_doc].annotation_collections[0].remove_duplicates()
        for ann in doc_coll.documents[i_doc].annotation_collections[0].annotations:
            if isinstance(ann, EntityAnnotation):
                n_ann_tokens += len(ann.content.replace("\n", " ").replace("\t", " ").split())

        n_tokens += len(doc.text.replace("\n", " ").replace("\t", " ").split())
    print()
    return len(doc_coll.documents), n_tokens, n_ann_tokens


def stat_ann_categories(paths: list[Path | str], max_n: int = -1, stats: dict | None = None, depth: int = 0, max_depth: int = 0) -> dict:
    """
    Collects and prints statistics about annotated categories in documents.
    """
    if stats is None:
        stats = {
            "n_tokens": 0,
            "n_annotated_tokens": 0,
            "n_annotations": 0,
            "n_documents": 0,
            "ann_types": {},
        }

    if not isinstance(paths, list):
        paths = [paths]

    for path in paths:
        if isinstance(path, str):
            path = Path(path)

        items = list(path.iterdir())
        if depth < max_depth:
            print("iterating", len(items), "items in", path)
            for item in tqdm(items):
                if item.is_dir():
                    stats = stat_ann_categories([item], stats=stats, depth=depth + 1)

        doc_coll: DocumentCollection | None = read_document_collection_from_folder(str(path))
        if doc_coll is not None:
            for doc in doc_coll.documents:
                for ann in doc.annotation_collections[0].annotations:
                    stats["ann_types"][ann.label] += stats["ann_types"].get(ann.label, 0) + 1
                stats["n_tokens"] += len(doc.text.replace("\n", " ").replace("\t", " ").split())
                stats["n_annotated_tokens"] += sum(
                    [len(ann.content.replace("\n", " ").replace("\t", " ").split()) for ann in doc.annotation_collections[0].annotations if isinstance(ann, EntityAnnotation)]
                )
                stats["n_annotations"] += len(doc.annotation_collections[0].annotations)
                stats["n_documents"] += 1

    if depth == 0:
        # remove 'text_spurious' from the statistics
        if "text_spurious" in stats["ann_types"]:
            del stats["ann_types"]["text_spurious"]

        df = pd.DataFrame(stats["ann_types"].items(), columns=["label", "count"])
        df["percentage"] = 100 * df["count"] / df["count"].sum()
        df = df.sort_values("count", ascending=False)
        del stats["ann_types"]
        print(df)
        print(stats)
        print("Percentage of annotated tokens: ", 100 * stats["n_annotated_tokens"] / stats["n_tokens"])

    return stats



from .agreement_types import Agreement, FragmentAgreement, MucCollection, MucTable
from .eval_src import (
    compare_annotations,
    compare_folders,
    compare_many_folders,
    compare_pairs_of_folders,
    count_entities_of_folder,
    count_entities_of_many_folders,
    count_tokens_and_anns,
    create_union_gold_candidate_sets,
    stat_ann_categories,
)
